// Copyright (C) 2000, Luca Padovani <luca.padovani@cs.unibo.it>.
// 
// This file is part of GtkMathView, a Gtk widget for MathML.
// 
// GtkMathView is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// GtkMathView is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with GtkMathView; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// 
// For details, see the GtkMathView World-Wide-Web page,
// http://cs.unibo.it/~lpadovan/mml-widget, or send a mail to
// <luca.padovani@cs.unibo.it>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <assert.h>

#include "globals.hh"
#include "Iterator.hh"
#include "CharMapper.hh"
#include "MathMLElement.hh"
#include "MathMLStretchyCharNode.hh"

MathMLStretchyCharNode::MathMLStretchyCharNode(Char ch) : MathMLCharNode(ch)
{
  stretch = charMapper.GetStretch(ch);
  assert(stretch != STRETCH_NO);
  // The basic idea is that a stretchable char behaves exactly as an unstretchable
  // one as long as the layout is not allocated (i.e.layout.simple == -1 && layout.n == 0).
  layout.simple = -1;
  layout.n = 0;
}

MathMLStretchyCharNode::~MathMLStretchyCharNode()
{
}

void
MathMLStretchyCharNode::Fontify(const FontAttributes& fa)
{
  fChar = charMapper.FontifyChar(fa, ch, true);
  if (fChar == NULL) MathMLCharNode::Fontify(fa);
}

void
MathMLStretchyCharNode::SetDefaultLargeGlyph(bool large)
{
  if (fChar == NULL || fChar->charMap == NULL || fChar->charMap->GetStretch() == STRETCH_NO) return;
  assert(fChar->charMap != NULL);
  fChar->nch = fChar->charMap->Map(ch, large);
}

void
MathMLStretchyCharNode::DoMinimumLayout()
{
  // WARNING: here we must be sure that the trivial char layout (without
  // stretchying capabilities) yields the smallest char. This is case
  // only if the charMap has at least one simple char, and if it has more,
  // then the first is the smallest one. Be careful with the map!
  layout.simple = -1;
  layout.n = 0;
  MathMLCharNode::DoLayout();
}

void
MathMLStretchyCharNode::DoVerticalStretchyLayout(scaled desiredSize,
						 float ascentR,
						 scaled axis,
						 bool strict)
{
  if (fChar == NULL) return;

  DoVerticalStretchyLayoutAux(desiredSize, strict);

  box.width = layout.charBox.width;
  scaled height = layout.charBox.GetHeight();
  box.ascent = float2sp(height * ascentR) + axis;
  box.descent = height - box.ascent;

#if 0
  printf("height %d ratio %f ascent %d descent %d\n", sp2ipx(height), ascentR, sp2ipx(box.ascent),
	 sp2ipx(box.descent));
#endif
}

void
MathMLStretchyCharNode::DoVerticalStretchyLayoutAux(scaled desiredSize, bool strict)
{
  assert(fChar != NULL);
  assert(fChar->font != NULL);

  const Font* font = fChar->font;
  const NativeChar* nch = fChar->charMap->stretchy.simple;

  layout.simple = -1;
  layout.n = 0;
  layout.charBox.Null();

  scaled best = 0;

  // first of all let's see if there is some single char large enough
  for (unsigned i = 0; i < MAX_SIMPLE_CHARS && nch[i] != 0; i++) {
    scaled charHeight = font->CharHeight(nch[i]);
    if ((scaledGeq(charHeight, desiredSize) &&
	(scaledAbs(desiredSize - charHeight) < scaledAbs(desiredSize - best)))) {
      // ok, the char is large enough and moreover it best approximate the
      // desired size (or it is the very first char found)
      layout.simple = i;
      best = charHeight;
    }
  }

  if (layout.simple >= 0) {
    font->CharBox(nch[layout.simple], layout.charBox);
    return;
  }

  // we didn't succeeded, so we look for a suitable compound char
  nch = fChar->charMap->stretchy.compound;

  scaled size  = 0; // total size of the stretched char
  scaled width = 0; // total width of the stretched char

  if (nch[SC_FIRST] != 0) {
    font->CharBox(nch[SC_FIRST], layout.box[SC_FIRST]);
    size += layout.box[SC_FIRST].GetHeight();
    width = scaledMax(width, layout.box[SC_FIRST].width);
    layout.n++;
  }

  if (nch[SC_LAST] != 0) {
    font->CharBox(nch[SC_LAST], layout.box[SC_LAST]);
    size += layout.box[SC_LAST].GetHeight();
    width = scaledMax(width, layout.box[SC_LAST].width);
    layout.n++;
  }

  if (nch[SC_MIDDLE] != 0) {
    font->CharBox(nch[SC_MIDDLE], layout.box[SC_MIDDLE]);
    size += layout.box[SC_MIDDLE].GetHeight();
    width = scaledMax(width, layout.box[SC_MIDDLE].width);
    layout.n++;
  }

  if (scaledGeq(size, desiredSize)) {
    if (nch[SC_LAST] != 0) layout.charBox = layout.box[SC_LAST];
    else if (nch[SC_MIDDLE] != 0) layout.charBox = layout.box[SC_MIDDLE];
    else layout.charBox = layout.box[SC_FIRST];

    layout.charBox.ascent += size - layout.charBox.GetHeight();
    layout.charBox.width = width;

    return;
  }

  if (nch[SC_REPEAT] != 0) font->CharBox(nch[SC_REPEAT], layout.box[SC_REPEAT]);

  if (nch[SC_REPEAT] != 0 && layout.box[SC_REPEAT].GetHeight() > SP_EPSILON) {
    width = scaledMax(width, layout.box[SC_REPEAT].width);
    while (desiredSize - size > SP_EPSILON) {
      unsigned step = (nch[SC_MIDDLE] == 0) ? 1 : 2;
      layout.n += step;
      size += step * layout.box[SC_REPEAT].GetHeight();
    }
  }

  if      (nch[SC_LAST] != 0) layout.charBox = layout.box[SC_LAST];
  else if (nch[SC_REPEAT] != 0) layout.charBox = layout.box[SC_REPEAT];
  else layout.charBox = layout.box[SC_FIRST];

  layout.charBox.ascent += size - layout.charBox.GetHeight();
  layout.charBox.width = width;
}

void
MathMLStretchyCharNode::DoHorizontalStretchyLayout(scaled desiredSize, bool strict)
{
  if (fChar == NULL) return;

  DoHorizontalStretchyLayoutAux(desiredSize, strict);
  box = layout.charBox;
}

void
MathMLStretchyCharNode::DoHorizontalStretchyLayoutAux(scaled desiredSize, bool strict)
{
  assert(fChar != NULL);
  assert(fChar->font != NULL);

  const Font* font = fChar->font;
  const NativeChar* nch = fChar->charMap->stretchy.simple;

  layout.simple = -1;
  layout.n = 0;
  layout.charBox.Null();

  scaled best = 0;

  // first of all let's see if there is some single char large enough
  for (unsigned i = 0; i < MAX_SIMPLE_CHARS && nch[i] != 0; i++) {
    scaled charWidth = font->CharWidth(nch[i]);
    if (scaledGeq(charWidth, desiredSize) &&
	(scaledAbs(desiredSize - charWidth) < scaledAbs(desiredSize - best))) {
      // ok, the char is large enough and moreover it best approximate the
      // desired size (or it is the very first char found)
      layout.simple = i;
      best = charWidth;
    }
  }

  if (layout.simple >= 0) {
    font->CharBox(nch[layout.simple], layout.charBox);
    return;
  }

  nch = fChar->charMap->stretchy.compound;

  if (nch[SC_FIRST] != 0) {
    font->CharBox(nch[SC_FIRST], layout.box[SC_FIRST]);
    layout.charBox.Append(layout.box[SC_FIRST]);
    layout.n++;
  }

  if (nch[SC_LAST] != 0) {
    font->CharBox(nch[SC_LAST], layout.box[SC_LAST]);
    layout.charBox.Append(layout.box[SC_LAST]);
    layout.n++;
  }

  if (nch[SC_MIDDLE] != 0) {
    font->CharBox(nch[SC_MIDDLE], layout.box[SC_MIDDLE]);
    layout.charBox.Append(layout.box[SC_MIDDLE]);
    layout.n++;
    if (scaledGeq(layout.charBox.width, desiredSize)) {
      return;
    }
  }

  if (scaledGeq(layout.charBox.width, desiredSize)) {
    return;
  }

  if (nch[SC_REPEAT] != 0) font->CharBox(nch[SC_REPEAT], layout.box[SC_REPEAT]);

  if (nch[SC_REPEAT] != 0 && layout.box[SC_REPEAT].width > SP_EPSILON) {
    while (desiredSize - layout.charBox.width > SP_EPSILON) {
      unsigned step = (nch[SC_MIDDLE] == 0) ? 1 : 2;
      layout.n += step;
      layout.charBox.Append(layout.box[SC_REPEAT]);
      if (step == 2)
	layout.charBox.Append(layout.box[SC_REPEAT]);
    }
  }
}

void
MathMLStretchyCharNode::Render(const DrawingArea& area)
{
  if (!HasDirtyChildren()) return;

  if (fChar == NULL || fChar->charMap->GetStretch() == STRETCH_NO ||
      (layout.simple < 0 && layout.n == 0)) MathMLCharNode::Render(area);
  else {
    assert(GetParent() != NULL);
    assert(GetParent()->IsToken());

    GraphicsContext* gc = GetParent()->GetGraphicsContext();
    assert(gc != NULL);

    if (fChar->charMap->GetStretch() == STRETCH_VERTICAL)
      RenderVerticalStretchyChar(area, gc, GetX(), GetY() + box.descent - layout.charBox.descent);
    else
      RenderHorizontalStretchyChar(area, gc, GetX(), GetY() + box.descent - layout.charBox.descent);
  }

  ResetDirty();
}

void
MathMLStretchyCharNode::RenderVerticalStretchyChar(const DrawingArea& area,
						   const GraphicsContext* gc, scaled x, scaled y)
{
  assert(gc != NULL);
  assert(fChar != NULL);
  assert(fChar->font != NULL);

  const Font* font = fChar->font;

  if (layout.simple >= 0) {
    area.DrawNativeChar(gc, font, x, y, fChar->charMap->stretchy.simple[layout.simple]);
    return;
  }

  const NativeChar* nch = fChar->charMap->stretchy.compound;

  unsigned nRepeat = layout.n;

  if (nch[SC_FIRST] != 0) nRepeat--;
  if (nch[SC_LAST] != 0) nRepeat--;
  if (nch[SC_MIDDLE] != 0 && nRepeat > 0) nRepeat--;

  if (nch[SC_LAST] != 0) {
    area.DrawNativeChar(gc, font, x, y, nch[SC_LAST]);
    y -= layout.box[SC_LAST].ascent;
  }

  if (nch[SC_MIDDLE] != 0) {  
    assert((nRepeat % 2) == 0);

    for (unsigned i = 0; i < nRepeat / 2; i++) {
      y -= layout.box[SC_REPEAT].descent;
      area.DrawNativeChar(gc, font, x, y, nch[SC_REPEAT]);
      y -= layout.box[SC_REPEAT].ascent;
    }

    y -= layout.box[SC_MIDDLE].descent;
    area.DrawNativeChar(gc, font, x, y, nch[SC_MIDDLE]);
    y -= layout.box[SC_MIDDLE].ascent;

    for (unsigned i = 0; i < nRepeat / 2; i++) {
      y -= layout.box[SC_REPEAT].descent;
      area.DrawNativeChar(gc, font, x, y, nch[SC_REPEAT]);
      y -= layout.box[SC_REPEAT].ascent;
    }
  } else {
    for (unsigned i = 0; i < nRepeat; i++) {
      y -= layout.box[SC_REPEAT].descent;
      area.DrawNativeChar(gc, font, x, y, nch[SC_REPEAT]);
      y -= layout.box[SC_REPEAT].ascent;
    }
  }

  if (nch[SC_FIRST] != 0) {
    y -= layout.box[SC_FIRST].descent;
    area.DrawNativeChar(gc, font, x, y, nch[SC_FIRST]);
  }
}

void
MathMLStretchyCharNode::RenderHorizontalStretchyChar(const DrawingArea& area,
						     const GraphicsContext* gc, scaled x, scaled y)
{
  assert(gc != NULL);
  assert(fChar != NULL);
  assert(fChar->font != NULL);

  const Font* font = fChar->font;

  if (layout.simple >= 0) {
    area.DrawNativeChar(gc, font, x, y, fChar->charMap->stretchy.simple[layout.simple]);
    return;
  }

  const NativeChar* nch = fChar->charMap->stretchy.compound;

  unsigned nRepeat = layout.n;

  if (nch[SC_FIRST] != 0) nRepeat--;
  if (nch[SC_LAST] != 0) nRepeat--;
  if (nch[SC_MIDDLE] != 0 && nRepeat > 0) nRepeat--;

  if (nch[SC_FIRST] != 0) {
    area.DrawNativeChar(gc, font, x, y, nch[SC_FIRST]);
    x += layout.box[SC_FIRST].width;
  }

  if (nch[SC_MIDDLE] != 0) {  
    assert((nRepeat % 2) == 0);

    for (unsigned i = 0; i < nRepeat / 2; i++) {
      area.DrawNativeChar(gc, font, x, y, nch[SC_REPEAT]);
      x += layout.box[SC_REPEAT].width;
    }

    area.DrawNativeChar(gc, font, x, y, nch[SC_MIDDLE]);
    x += layout.box[SC_MIDDLE].width;

    for (unsigned i = 0; i < nRepeat / 2; i++) {
      area.DrawNativeChar(gc, font, x, y, nch[SC_REPEAT]);
      x += layout.box[SC_REPEAT].width;
    }
  } else {
    for (unsigned i = 0; i < nRepeat; i++) {
      area.DrawNativeChar(gc, font, x, y, nch[SC_REPEAT]);
      x += layout.box[SC_REPEAT].width;
    }
  }

  if (nch[SC_LAST] != 0) {
    area.DrawNativeChar(gc, font, x, y, nch[SC_LAST]);
  }
}

bool
MathMLStretchyCharNode::IsStretchyChar() const
{
  return true;
}
