// Copyright (C) 2000, Luca Padovani <luca.padovani@cs.unibo.it>.
// 
// This file is part of GtkMathView, a Gtk widget for MathML.
// 
// GtkMathView is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// GtkMathView is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with GtkMathView; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// 
// For details, see the GtkMathView World-Wide-Web page,
// http://cs.unibo.it/~lpadovan/mml-widget, or send a mail to
// <luca.padovani@cs.unibo.it>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <wctype.h>

#include "String.hh"
#include "Parser.hh"
#include "MathML.hh"
#include "globals.hh"
#include "stringAux.hh"
#include "EntitiesTable.hh"
#include "allocTextNode.hh"
#include "MathMLParseFile.hh"

bool verbose = false;

MathMLParser::MathMLParser(const char* s)
{
  fileName = s;
  doc = NULL;
}

MathMLParser::~MathMLParser()
{
}

MathMLDocument*
MathMLParser::Parse()
{
  doc = MathMLParseFile(fileName, false);
  if (doc == NULL) return NULL;
  if (xmlDocGetRootElement(doc) == NULL) return NULL;

  MathMLDocument* document = new MathMLDocument();
  MathMLizeNode(xmlDocGetRootElement(doc), document);

  return document;
}

void
MathMLParser::Dump(FILE* f)
{
  xmlDocDump(f, doc);
}

// MathMLizeNode: main parsing function to recognize a MathML element.
// node -> source node in the DOM tree
// parent -> parent (parsed) node
void
MathMLParser::MathMLizeNode(xmlNodePtr node, MathMLContainerElement* parent)
{
  assert(node != NULL);
  assert(parent != NULL);

  // TODO: namespaces
  //  if (node->ns != NULL && xmlStrcmp(node->ns->prefix, XMLSTRING(MATHML_NS_PREFIX))) {
  //  if (verbose) fprintf(stderr, "WARNING: skipping element with namespace `%s'\n", node->ns->prefix);
  //  return;
  //}

  if (node->type != XML_ELEMENT_NODE) {
    if (verbose) fprintf(stderr, "WARNING: skipping unrecognized node (type %d)\n", node->type);
    return;
  }

  TagId tag = TagIdOfName(CSTRING(node->name));
  MathMLElement* elem = NULL;

  switch (tag) {
  case TAG_MATH:
    elem = new MathMLmathElement();
    break;
  case TAG_MI:
  case TAG_MN:
    elem = new MathMLTokenElement(tag);
    break;
  case TAG_MO:
    elem = new MathMLOperatorElement();
    break;
  case TAG_MTEXT:
    elem = new MathMLTextElement();
    break;
  case TAG_MSPACE:
    elem = new MathMLSpaceElement();
    break;
  case TAG_MS:
    elem = new MathMLStringLitElement();
    break;
  case TAG_MROW:
    elem = new MathMLRowElement();
    break;
  case TAG_MFRAC:
    elem = new MathMLFractionElement();
    break;
  case TAG_MSQRT:
  case TAG_MROOT:
    elem = new MathMLRadicalElement(tag);
    break;
  case TAG_MSTYLE:
    elem = new MathMLStyleElement();
    break;
  case TAG_MERROR:
    elem = new MathMLErrorElement();
    break;
  case TAG_MPADDED:
    elem = new MathMLPaddedElement();
    break;
  case TAG_MPHANTOM:
    elem = new MathMLPhantomElement();
    break;
  case TAG_MFENCED:
    elem = new MathMLFencedElement();
    break;
  case TAG_MSUB:
  case TAG_MSUP:
  case TAG_MSUBSUP:
    elem = new MathMLScriptElement(tag);
    break;
  case TAG_MUNDER:
  case TAG_MOVER:
  case TAG_MUNDEROVER:
    elem = new MathMLUnderOverElement(tag);
    break;
  case TAG_MMULTISCRIPTS:
    elem = new MathMLMultiScriptsElement();
    break;
  case TAG_MPRESCRIPTS:
  case TAG_NONE:
    if (parent != NULL && parent->IsA() == TAG_MMULTISCRIPTS)
      elem = new MathMLElement(tag);
    break;
  case TAG_MTABLE:
    elem = new MathMLTableElement();
    break;
  case TAG_MTR:
    if (parent != NULL && parent->IsA() == TAG_MTABLE)
      elem = new MathMLTableRowElement();
    break;
  case TAG_MLABELEDTR:
    if (parent != NULL && parent->IsA() == TAG_MTABLE)
      elem = new MathMLLabeledRowElement();
    break;
  case TAG_MTD:
    if (parent != NULL && (parent->IsA() == TAG_MTR || parent->IsA() == TAG_MLABELEDTR || parent->IsA() == TAG_MTABLE))
      elem = new MathMLTableCellElement();
    break;
  case TAG_MALIGNGROUP:
    elem = new MathMLAlignGroupElement();
    break;
  case TAG_MALIGNMARK:
    elem = new MathMLAlignMarkElement();
    break;
  case TAG_MACTION:
    elem = new MathMLActionElement();
    break;
  case TAG_SEMANTICS:
    elem = new MathMLSemanticsElement();
    break;
  case TAG_MCHAR:
    if (verbose)
      fprintf(stderr, "WARNING: `mchar' is valid inside token elements only (ignored)\n");
    break;
  case TAG_NOTVALID:
  default:
    break;
  }

  if (elem != NULL) {
    MathMLizeAttributes(node, elem);
    elem->SetParent(parent);
    parent->content.Append(elem);
  } else if (verbose) {
    if (tag != TAG_NOTVALID) {
      fprintf(stderr, "WARNING: `%s' is not allowed ", NameOfTagId(tag));
      if (parent == NULL) fprintf(stderr, "as root element");
      else fprintf(stderr, "inside a `%s' element", NameOfTagId(parent->IsA()));
      fprintf(stderr, " (ignored)\n");
    } else {
      fprintf(stderr, "WARNING: unrecognized tag `%s' (ignored)\n", node->name);
    }
  }

  if      (elem == NULL) MathMLizeContainerContent(node, parent);
  else if (elem->IsContainer()) MathMLizeContainerContent(node, TO_CONTAINER(elem));
  else if (elem->IsToken()) MathMLizeTokenContent(node, TO_TOKEN(elem));
}

// MathMLizeAttributes: associate valid attributes to the node
// node -> DOM node containing the attributes
// parent -> parsed node the attributes belongs to
void
MathMLParser::MathMLizeAttributes(xmlNodePtr node, MathMLElement* parent)
{
  assert(node != NULL);
  assert(parent !=NULL);

  xmlAttrPtr p = node->properties;

  while (p != NULL) {
    if (p->ns != NULL &&
	xmlStrcmp(p->ns->prefix, XMLSTRING(MATHML_NS_PREFIX))
	&& xmlStrcmp(p->ns->prefix, XMLSTRING(XLINK_NS_PREFIX))
#ifdef HELM_PROJECT
	&& xmlStrcmp(p->ns->prefix, XMLSTRING(HELM_NS_PREFIX))
#endif
	) {
      if (verbose)
	fprintf(stderr, "WARNING: unsupported attribute `%s:%s' (ignored)\n", p->ns->prefix, p->name);
    } else {
      AttributeId id;

      if (p->ns != NULL)
	id = AttributeIdOfName(CSTRING(p->ns->prefix), CSTRING(p->name));
      else
	id = AttributeIdOfName(CSTRING(p->name));

      const AttributeSignature* aSignature =
	(id != ATTR_NOTVALID) ? parent->GetAttributeSignature(id) : NULL;
      
      if (id == ATTR_NOTVALID || (aSignature == NULL && parent->IsA() != TAG_MSTYLE)) {
	// this attribute is not supported by the current element
	if (verbose)
	  fprintf(stderr, "WARNING: unsupported attribute `%s' (ignored)\n", p->name);
      } else {
#ifdef HAVE_LIBXML2
	xmlChar* aValue = xmlNodeListGetString(doc, p->children, 1);
#else
	xmlChar* aValue = xmlGetProp(node, p->name);
#endif
	if (aValue != NULL) {
	  const String* sValue = allocString(aValue);	
	  MathMLAttribute* attribute = new MathMLAttribute(id, sValue);
	  parent->attributes.Append(attribute);
	}
      }
    }

    p = p->next;
  }
}

void
MathMLParser::MathMLizeContainerContent(xmlNodePtr node, MathMLContainerElement* parent)
{
  assert(node != NULL);
  assert(parent != NULL);

#ifdef HAVE_LIBXML2
  xmlNodePtr p = node->children;
#else
  xmlNodePtr p = node->childs;
#endif

  while (p != NULL) {
    if (xmlNodeIsText(p)) {
      String* text = allocString(xmlNodeGetContent(p));
      assert(text != NULL);

      text->CollapseSpaces();
      text->TrimSpacesLeft();
      text->TrimSpacesRight();

      if (text->GetLength() > 0) {
	if (verbose)
	  fprintf(stderr, "WARNING: text inside container element `%s' (ignored)\n", NameOfTagId(parent->IsA()));
      }

      delete text;
    } else
      MathMLizeNode(p, parent);

    p = p->next;
  }
}

void
MathMLParser::MathMLizeTokenContent(xmlNodePtr node, MathMLTokenElement* parent)
{
  assert(node != NULL);
  assert(parent != NULL);

#ifdef HAVE_LIBXML2
  xmlNodePtr p = node->children;
#else
  xmlNodePtr p = node->childs;
#endif

  bool     lastSpace = false;

  while (p != NULL) {
    if (xmlNodeIsText(p)) {
      // ok, we have a chunk of text
      String* s = allocString(xmlNodeGetContent(p));
      assert(s != NULL);

      // white-spaces are always collapsed...
      s->CollapseSpaces();

      // ...but spaces at the end or at the beginning are deleted only if this
      // is the very first chunk in the token.
      if (p->prev == NULL) s->TrimSpacesLeft();
      if (p->next == NULL) s->TrimSpacesRight();

      AtomizeTextContent(s, parent);
    } else if (p->type == XML_ENTITY_REF_NODE) {
      if (p->name != NULL) {
	String* s = entitiesTable.GetEntityContent(p->name);
	if (s == NULL) s = entitiesTable.GetErrorEntityContent();
	assert(s != NULL);

	AtomizeTextContent(s, parent);
      }
    } else if (p->type == XML_ELEMENT_NODE) {
      TagId tag = TagIdOfName(CSTRING(p->name));

      switch (tag) {
      case TAG_MCHAR:
	{
	  String* s = SubstituteMCharElement(p);
	  if (s != NULL) AtomizeTextContent(s, parent);
	}
	break;
      case TAG_MGLYPH:
	{
	  MathMLTextNode* node = SubstituteMGlyphElement(p);
	  if (node != NULL) {
	    node->SetParent(parent);
	    parent->content.Append(node);
	  }
	}
	break;
      case TAG_MALIGNMARK:
	{
	  // we cannot use MathMLizeNode beacuse it takes a container as
	  // parent element, while here parent is a token element
	  MathMLElement* elem = new MathMLAlignMarkElement();
	  MathMLizeAttributes(p, elem);
	  elem->SetParent(parent);
	  parent->content.Append(elem);
	}
	break;
      default:
	if (verbose)
	  fprintf(stderr, "WARNING: unacceptable element `%s' inside token (ignored)\n", p->name);
	break;
      }

      lastSpace = false;
    }
    
    p = p->next;
  }
}

// AtomizeTextContent: break a source string S into atomic nodes and insert them
// into the token element PARENT. Breakable spaces get DEFAULTID breakability
void
MathMLParser::AtomizeTextContent(const String* s, MathMLTokenElement* parent)
{
  assert(s != NULL);
  assert(parent != NULL);

  if (s->GetLength() == 0) {
    delete s;
    return;
  }

  unsigned i = 0;
  while (i < s->GetLength()) {
    MathMLTextNode* node = NULL;

    if (s->GetChar(i) == ' ' && parent->content.GetSize() > 0) {
      assert(parent->content.GetLast() != NULL);
      if (parent->content.GetLast()->IsText()) {
	MathMLTextNode* lastNode = TO_TEXT(parent->content.GetLast());
	assert(lastNode != NULL);

	if (!lastNode->HasExtraSpace()) {
	  node = lastNode;
	  node->AddExtraSpace();
	  i++;
	}
      }
    }

    if (node == NULL) {
      node = parseString(*s, i);
      assert(node != NULL);
      node->SetParent(parent);
      parent->content.Append(node);
    }
  }

  delete s;
}

String*
MathMLParser::SubstituteMCharElement(xmlNodePtr node)
{
  assert(node != NULL);

  xmlChar* entityName = xmlGetProp(node, XMLSTRING("name"));
  String* content = NULL;

  if (entityName == NULL) {
    if (verbose)
      fprintf(stderr, "WARNING: malformed `mchar' element (`name' attribute required)\n");

    content = allocString(XMLSTRING("?"));
  } else {
    content = entitiesTable.GetEntityContent(entityName);
    if (content == NULL) content = entitiesTable.GetErrorEntityContent();
  }

  assert(content != NULL);

  if (content->GetLength() == 0) {
    delete content;
    content = NULL;
  }

  return content;
}

MathMLTextNode*
MathMLParser::SubstituteMGlyphElement(xmlNodePtr node)
{
  assert(node != NULL);

  xmlChar* alt = xmlGetProp(node, XMLSTRING("alt"));
  xmlChar* fontFamily = xmlGetProp(node, XMLSTRING("fontfamily"));
  xmlChar* index = xmlGetProp(node, XMLSTRING("index"));

  if (alt == NULL || fontFamily == NULL || index == NULL ||
      *alt == '\0' || *fontFamily == '\0' || *index == '\0') {
    if (verbose)
      fprintf(stderr, "WARNING: malformed `mglyph' element (some required attribute is missing)\n");

    return new MathMLCharNode('?');
  }

  char* endPtr;
  unsigned nch = strtoul(CSTRING(index), &endPtr, 10);

  if (endPtr == NULL || *endPtr != '\0') {
    if (verbose)
      fprintf(stderr, "WARNING: malformed `mglyph' element (parsing error in `index' attribute)\n");

    nch = '?';
  }

  return new MathMLGlyphNode(CSTRING(alt), CSTRING(fontFamily), nch);
}
