/*
 * Copyright (C) 2000, Luca Padovani <luca.padovani@cs.unibo.it>.
 * 
 * This file is part of GtkMathView, a Gtk widget for MathML.
 * 
 * GtkMathView is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * GtkMathView is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with GtkMathView; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * For details, see the GtkMathView World-Wide-Web page,
 * http://cs.unibo.it/~lpadovan/mml-widget, or send a mail to
 * <luca.padovani@cs.unibo.it>
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "defs.h"

#include <gdk/gdkkeysyms.h>

#include <gtk/gtkmain.h>
#include <gtk/gtksignal.h>
#include <gtk/gtkdrawingarea.h>

#include "CCInterface.h"
#include "gtkmathview.h"

static void gtk_math_view_class_init(GtkMathViewClass*);
static void gtk_math_view_init(GtkMathView*);

/* GtkObject functions */

static void gtk_math_view_destroy(GtkObject*);

/* GtkWidget functions */

static gboolean gtk_math_view_configure_event(GtkWidget*, GdkEventConfigure*, GtkMathView*);
static gboolean gtk_math_view_expose_event(GtkWidget*, GdkEventExpose*, GtkMathView*);
static gboolean gtk_math_view_button_press_event(GtkWidget*, GdkEventButton*, GtkMathView*);
static gboolean gtk_math_view_button_release_event(GtkWidget*, GdkEventButton*, GtkMathView*);
static gboolean gtk_math_view_motion_notify_event(GtkWidget*, GdkEventMotion*, GtkMathView*);
static gboolean gtk_math_view_key_press_event(GtkWidget*, GdkEventKey*, GtkMathView*);

/* GtkMathView functions */

static void gtk_math_view_clicked(GtkMathView*, gpointer);
static void gtk_math_view_jump(GtkMathView*, const gchar*, gpointer);

/* auxiliary functions */

static void set_current_cursor(GtkMathView*, gint, gint);
static void setup_adjustment(GtkAdjustment*, gfloat, gfloat);
static void setup_adjustments(GtkMathView*);

/* Local data */

static GtkDrawingAreaClass* parent_class = NULL;
static guint clicked_signal = 0;
static guint jump_signal = 0;

/* widget implementation */

static void
paint_widget_area(GtkMathView* math_view, gint x, gint y, gint width, gint height)
{
  GtkWidget* widget;

  g_return_if_fail(math_view != NULL);
  g_return_if_fail(math_view->area != NULL);

  widget = math_view->area;

  gdk_draw_rectangle(math_view->pixmap, widget->style->white_gc, TRUE, x, y, width, height);
  render_document(math_view->interface,
		  (gint) (math_view->top_x / SCALED_POINTS_PER_PX) + x,
		  (gint) (math_view->top_y / SCALED_POINTS_PER_PX) + y,
		  width, height);
}

static void
paint_widget(GtkMathView* math_view)
{
  GtkWidget* widget;

  g_return_if_fail(math_view != NULL);
  g_return_if_fail(math_view->area != NULL);

  widget = math_view->area;

  paint_widget_area(math_view, 0, 0, widget->allocation.width, widget->allocation.height);
}

static void
hadjustment_value_changed(GtkAdjustment* adj, GtkMathView* math_view)
{
  g_return_if_fail(adj != NULL);
  g_return_if_fail(math_view != NULL);

  if (adj->value > adj->upper - adj->page_size) adj->value = adj->upper - adj->page_size;
  if (adj->value < adj->lower) adj->value = adj->lower;

  math_view->old_top_x = math_view->top_x;
  math_view->top_x = adj->value;

  if (math_view->old_top_x != math_view->top_x) {
#if 0
    gint change = abs(math_view->old_top_x - math_view->top_x);
    GtkWidget* widget = math_view->area;
    if (change < widget->allocation.width) {
      if (math_view->old_top_x < math_view->top_x) {
	// the window scrolled right
	gdk_draw_pixmap(math_view->pixmap,
			widget->style->white_gc,
			math_view->pixmap,
			change, 0,
			0, 0,
			widget->allocation.width - change,
			widget->allocation.height);

	paint_widget_area(math_view,
			  widget->allocation.width - change, 0,
			  change, widget->allocation.height);

	gtk_widget_draw(widget, NULL);
      } else {
	// the window scrolled left
	gdk_draw_pixmap(math_view->pixmap,
			widget->style->white_gc,
			math_view->pixmap,
			0, 0,
			change, 0,
			widget->allocation.width - change,
			widget->allocation.height);

	paint_widget_area(math_view,
			  0, 0,
			  change, widget->allocation.height);

	gtk_widget_draw(widget, NULL);
      }
    } else
#endif
      paint_widget(math_view);
  }
}

static void
vadjustment_value_changed(GtkAdjustment* adj, GtkMathView* math_view)
{
  g_return_if_fail(adj != NULL);
  g_return_if_fail(math_view != NULL);

  if (adj->value > adj->upper - adj->page_size) adj->value = adj->upper - adj->page_size;
  if (adj->value < adj->lower) adj->value = adj->lower;

  math_view->old_top_y = math_view->top_y;
  math_view->top_y = adj->value;

  if (math_view->old_top_y != math_view->top_y) {
#if 0
    gint change = abs(math_view->old_top_y - math_view->top_y);
    GtkWidget* widget = math_view->area;
    if (change < widget->allocation.height) {
      if (math_view->old_top_y < math_view->top_y) {
	// the window scrolled down
	gdk_draw_pixmap(math_view->pixmap,
			widget->style->white_gc,
			math_view->pixmap,
			0, change,
			0, 0,
			widget->allocation.width,
			widget->allocation.height - change);

	paint_widget_area(math_view,
			  0, widget->allocation.height - change,
			  widget->allocation.width, change);

	gtk_widget_draw(widget, NULL);
      } else {
	// the window scrolled up
	gdk_draw_pixmap(math_view->pixmap,
			widget->style->white_gc,
			math_view->pixmap,
			0, 0,
			0, change,
			widget->allocation.width,
			widget->allocation.height - change);

	paint_widget_area(math_view,
			  0, 0,
			  widget->allocation.width, change);

	gtk_widget_draw(widget, NULL);
      }
    } else
#endif
      paint_widget(math_view);
  }
}

GtkType
gtk_math_view_get_type()
{
  static guint math_view_type = 0;

  if (math_view_type == 0) {
    GtkTypeInfo math_view_info = {
      "GtkMathView",
      sizeof(GtkMathView),
      sizeof(GtkMathViewClass),
      (GtkClassInitFunc) gtk_math_view_class_init,
      (GtkObjectInitFunc) gtk_math_view_init,
      (GtkArgSetFunc) NULL,
      (GtkArgGetFunc) NULL
    };

    math_view_type = gtk_type_unique(gtk_event_box_get_type(), &math_view_info);
  }

  return math_view_type;
}

static void
gtk_math_view_class_init(GtkMathViewClass* klass)
{
  GtkObjectClass* object_class = (GtkObjectClass*) klass;
  GtkWidgetClass* widget_class = (GtkWidgetClass*) klass;

  klass->normal_cursor = gdk_cursor_new(GDK_TOP_LEFT_ARROW);
  klass->link_cursor = gdk_cursor_new(GDK_HAND2);

  klass->clicked = gtk_math_view_clicked;
  klass->jump = gtk_math_view_jump;

  klass->set_scroll_adjustments = gtk_math_view_set_adjustments;

  parent_class = gtk_type_class(gtk_event_box_get_type());

  object_class->destroy = gtk_math_view_destroy;

  widget_class->set_scroll_adjustments_signal =
    gtk_signal_new("set_scroll_adjustments",
		   GTK_RUN_LAST,
		   object_class->type,
		   GTK_SIGNAL_OFFSET (GtkMathViewClass, set_scroll_adjustments),
		   gtk_marshal_NONE__POINTER_POINTER,
		   GTK_TYPE_NONE, 2, GTK_TYPE_ADJUSTMENT, GTK_TYPE_ADJUSTMENT);

  clicked_signal =
    gtk_signal_new("clicked",
		   GTK_RUN_FIRST,
		   object_class->type,
		   GTK_SIGNAL_OFFSET (GtkMathViewClass, clicked),
		   gtk_marshal_NONE__NONE,
		   GTK_TYPE_NONE, 0);

  gtk_object_class_add_signals(object_class, &clicked_signal, 1);

  jump_signal =
    gtk_signal_new("jump",
		   GTK_RUN_FIRST,
		   object_class->type,
		   GTK_SIGNAL_OFFSET (GtkMathViewClass, jump),
		   gtk_marshal_NONE__STRING,
		   GTK_TYPE_NONE, 1, GTK_TYPE_STRING);

  gtk_object_class_add_signals(object_class, &jump_signal, 1);
  
  init_global_data();
}

static void
gtk_math_view_init(GtkMathView* math_view)
{
  math_view->frame       = NULL;
  math_view->area        = NULL;
  math_view->pixmap      = NULL;
  math_view->vadjustment = NULL;
  math_view->hadjustment = NULL;
  math_view->interface   = NULL;
  math_view->select      = FALSE;
#ifdef HELM_PROJECT
  math_view->mode        = MATH_VIEW_SELECT_CONTENT;
#endif
}

GtkWidget*
gtk_math_view_new(GtkAdjustment* hadj, GtkAdjustment* vadj)
{
  GtkMathView* math_view = gtk_type_new(gtk_math_view_get_type());
  
  g_return_val_if_fail(math_view != NULL, NULL);

  math_view->area = gtk_drawing_area_new();
  gtk_widget_show(math_view->area);

  GTK_WIDGET_SET_FLAGS(GTK_WIDGET(math_view->area), GTK_CAN_FOCUS);

  math_view->frame = gtk_frame_new(NULL);
  gtk_widget_show(math_view->frame);
  gtk_container_add(GTK_CONTAINER(math_view->frame), math_view->area);
  gtk_container_add(GTK_CONTAINER(math_view), math_view->frame);

  gtk_frame_set_shadow_type(GTK_FRAME(math_view->frame), GTK_SHADOW_IN);

  gtk_signal_connect(GTK_OBJECT(math_view->area), "configure_event",
		     GTK_SIGNAL_FUNC(gtk_math_view_configure_event), math_view);

  gtk_signal_connect(GTK_OBJECT(math_view->area), "expose_event",
		     GTK_SIGNAL_FUNC(gtk_math_view_expose_event), math_view);

  gtk_signal_connect(GTK_OBJECT(math_view->area), "button_press_event",
		     GTK_SIGNAL_FUNC(gtk_math_view_button_press_event), math_view);

  gtk_signal_connect(GTK_OBJECT(math_view->area), "button_release_event",
		     GTK_SIGNAL_FUNC(gtk_math_view_button_release_event), math_view);

  gtk_signal_connect(GTK_OBJECT(math_view->area), "motion_notify_event",
		     GTK_SIGNAL_FUNC(gtk_math_view_motion_notify_event), math_view);

  gtk_signal_connect(GTK_OBJECT(math_view->area), "key_press_event",
		     GTK_SIGNAL_FUNC(gtk_math_view_key_press_event), math_view);

  math_view->top_x = math_view->top_y = 0;
  math_view->old_top_x = math_view->old_top_y = 0;
  
  math_view->interface = create_interface(math_view);

  math_view->hadjustment = NULL;
  math_view->vadjustment = NULL;

  gtk_widget_add_events(GTK_WIDGET(math_view->area),
			GDK_BUTTON_PRESS_MASK
			| GDK_BUTTON_RELEASE_MASK
			| GDK_POINTER_MOTION_MASK
			| GDK_KEY_PRESS_MASK);

  return GTK_WIDGET(math_view);
}

static void
gtk_math_view_destroy(GtkObject* object)
{
  GtkMathView* math_view;

  g_return_if_fail(object != NULL);
  g_return_if_fail(GTK_IS_MATH_VIEW(object));

  math_view = GTK_MATH_VIEW(object);

  destroy_interface(math_view->interface);

  if (math_view->frame != NULL)
    gtk_object_unref(GTK_OBJECT(math_view->frame));

  if (math_view->area != NULL)
    gtk_object_unref(GTK_OBJECT(math_view->area));

  if (math_view->hadjustment != NULL)
    gtk_object_unref(GTK_OBJECT(math_view->hadjustment));

  if (math_view->vadjustment != NULL)
    gtk_object_unref(GTK_OBJECT(math_view->vadjustment));

  if (GTK_OBJECT_CLASS(parent_class)->destroy != NULL)
    (*GTK_OBJECT_CLASS(parent_class)->destroy)(object);
}

static gboolean
gtk_math_view_key_press_event(GtkWidget* widget,
			      GdkEventKey* event,
			      GtkMathView* math_view)
{
  g_return_val_if_fail(widget != NULL, FALSE);
  g_return_val_if_fail(event != NULL, FALSE);
  g_return_val_if_fail(math_view != NULL, FALSE);

  if (event->type != GDK_KEY_PRESS) return FALSE;

  printf("key press %d\n", event->time);

  switch (event->keyval) {
  case GDK_Up:
  case GDK_KP_Up:
    math_view->vadjustment->value -= math_view->vadjustment->step_increment;
    gtk_adjustment_value_changed(math_view->vadjustment);
    break;
  case GDK_Down:
  case GDK_KP_Down:
    math_view->vadjustment->value += math_view->vadjustment->step_increment;
    gtk_adjustment_value_changed(math_view->vadjustment);
    break;
  case GDK_Left:
  case GDK_KP_Left:
    math_view->hadjustment->value -= math_view->hadjustment->step_increment;
    gtk_adjustment_value_changed(math_view->hadjustment);
    break;
  case GDK_Right:
  case GDK_KP_Right:
    math_view->hadjustment->value += math_view->hadjustment->step_increment;
    gtk_adjustment_value_changed(math_view->hadjustment);
    break;
  case GDK_Page_Up:
  case GDK_KP_Page_Up:
    math_view->vadjustment->value -= math_view->vadjustment->page_increment;
    gtk_adjustment_value_changed(math_view->vadjustment);
    break;
  case GDK_Page_Down:
  case GDK_KP_Page_Down:
    math_view->vadjustment->value += math_view->vadjustment->page_increment;
    gtk_adjustment_value_changed(math_view->vadjustment);
    break;
  case GDK_Home:
  case GDK_KP_Home:
    math_view->vadjustment->value = math_view->vadjustment->lower;
    gtk_adjustment_value_changed(math_view->vadjustment);
    break;
  case GDK_End:
  case GDK_KP_End:
    math_view->vadjustment->value = math_view->vadjustment->upper;
    gtk_adjustment_value_changed(math_view->vadjustment);
    break;
  default:
    return FALSE;
  }

  return TRUE;
}

static gint
gtk_math_view_button_press_event(GtkWidget* widget,
				 GdkEventButton* event,
				 GtkMathView* math_view)
{
  g_return_val_if_fail(widget != NULL, FALSE);
  g_return_val_if_fail(event != NULL, FALSE);
  g_return_val_if_fail(math_view != NULL, FALSE);

  if (event->button == 1) {
    set_selection_first(math_view->interface, event->x, event->y);
    math_view->button_press = TRUE;
    math_view->select = FALSE;
  }

  return FALSE;
}

static gint
gtk_math_view_button_release_event(GtkWidget* widget,
				   GdkEventButton* event,
				   GtkMathView* math_view)
{
  g_return_val_if_fail(widget != NULL, FALSE);
  g_return_val_if_fail(event != NULL, FALSE);
  g_return_val_if_fail(math_view != NULL, FALSE);

  if (event->button == 1) {
    if (math_view->button_press == TRUE && math_view->select == FALSE) {      
      const char* link = NULL;

      gtk_signal_emit(GTK_OBJECT(math_view), clicked_signal);

      link = get_hyperlink(math_view->interface, event->x, event->y);
      if (link != NULL) gtk_signal_emit(GTK_OBJECT(math_view), jump_signal, link);
    }

    math_view->button_press = math_view->select = FALSE;
  }

  return FALSE;
}

static gint
gtk_math_view_motion_notify_event(GtkWidget* widget,
				  GdkEventMotion* event,
				  GtkMathView* math_view)
{
  g_return_val_if_fail(widget != NULL, FALSE);
  g_return_val_if_fail(event != NULL, FALSE);
  g_return_val_if_fail(math_view != NULL, FALSE);

  if (event->x < 0) {
    math_view->hadjustment->value -= math_view->hadjustment->step_increment;
    gtk_adjustment_value_changed(math_view->hadjustment);
  } else if (event->x > widget->allocation.width) {
    math_view->hadjustment->value += math_view->hadjustment->step_increment;
    gtk_adjustment_value_changed(math_view->hadjustment);
  }

  if (event->y < 0) {
    math_view->vadjustment->value -= math_view->vadjustment->step_increment;
    gtk_adjustment_value_changed(math_view->vadjustment);
  } else if (event->y > widget->allocation.height) {
    math_view->vadjustment->value += math_view->vadjustment->step_increment;
    gtk_adjustment_value_changed(math_view->vadjustment);
  }

  math_view->select = math_view->button_press;

  if (math_view->select == TRUE)
    set_selection_last(math_view->interface, event->x, event->y);
  else
    set_current_cursor(math_view, event->x, event->y);

  return FALSE;
}

static void
set_current_cursor(GtkMathView* math_view, gint x, gint y)
{
  GtkMathViewClass* klass;

  g_return_if_fail(math_view != NULL);
  
  klass = (GtkMathViewClass*) gtk_type_class(gtk_math_view_get_type());
  g_return_if_fail(klass != NULL);
  
  if (is_on_link(math_view->interface, x, y))
    gdk_window_set_cursor(GTK_WIDGET(math_view)->window, klass->link_cursor);
  else
    gdk_window_set_cursor(GTK_WIDGET(math_view)->window, klass->normal_cursor);
}

static gint
gtk_math_view_configure_event(GtkWidget* widget,
			      GdkEventConfigure* event,
			      GtkMathView* math_view)
{
  gboolean do_layout;

  g_return_val_if_fail(widget != NULL, FALSE);
  g_return_val_if_fail(event != NULL, FALSE);
  g_return_val_if_fail(math_view != NULL, FALSE);

  if (math_view->pixmap != NULL) {
    gdk_pixmap_unref(math_view->pixmap);

    /* the actual layout is done onyl if the horizontal width of the
     * widget is changed, because it is the only parameter which
     * affects layout when this event occurs
     */
    do_layout = math_view->old_width != event->width;
  } else
    do_layout = TRUE;

  math_view->old_width = event->width;
  math_view->pixmap = gdk_pixmap_new(widget->window,
				     event->width,
				     event->height,
				     -1);

  if (do_layout) layout_document(math_view->interface, event->width);

  setup_adjustments(math_view);

  paint_widget(math_view);

  return TRUE;
}

static gint
gtk_math_view_expose_event(GtkWidget* widget,
			   GdkEventExpose* event,
			   GtkMathView* math_view)
{
  g_return_val_if_fail(widget != NULL, FALSE);
  g_return_val_if_fail(event != NULL, FALSE);
  g_return_val_if_fail(math_view != NULL, FALSE);

  gdk_draw_pixmap(widget->window,
		  widget->style->fg_gc[GTK_WIDGET_STATE(widget)],
		  math_view->pixmap,
		  event->area.x, event->area.y,
		  event->area.x, event->area.y,
		  event->area.width, event->area.height);

  return FALSE;
}

static void
setup_adjustment(GtkAdjustment* adj, gfloat size, gfloat page_size)
{
  g_return_if_fail(adj != NULL);

  adj->lower = 0.0;
  adj->page_size = page_size;
  adj->step_increment = 10 * SCALED_POINTS_PER_PX;
  adj->page_increment = page_size;
  adj->upper = size + 10 * SCALED_POINTS_PER_PX;
  if (adj->upper < 0) adj->upper = 0.0;

  if (page_size < adj->upper && adj->value + page_size > adj->upper) {
    adj->value = adj->upper - page_size;
    gtk_adjustment_value_changed(adj);
  }

  gtk_adjustment_changed(adj);
  gtk_adjustment_set_value(adj, 0.0);
}

static void
setup_adjustments(GtkMathView* math_view)
{
  g_return_if_fail(math_view != NULL);
  g_return_if_fail(math_view->area != NULL);

  math_view->old_top_x = math_view->old_top_y = math_view->top_x = math_view->top_y = 0;

  if (math_view->hadjustment != NULL)
    setup_adjustment(math_view->hadjustment,
		     get_document_width(math_view->interface) * SCALED_POINTS_PER_PX,
		     math_view->area->allocation.width * SCALED_POINTS_PER_PX);

  if (math_view->vadjustment != NULL)
    setup_adjustment(math_view->vadjustment,
		     get_document_height(math_view->interface) * SCALED_POINTS_PER_PX,
		     math_view->area->allocation.height * SCALED_POINTS_PER_PX);
}

gboolean
gtk_math_view_load(GtkMathView* math_view, const gchar* name)
{
  gboolean res;

  g_return_val_if_fail(math_view != NULL, FALSE);
  g_return_val_if_fail(math_view->area != NULL, FALSE);
  g_return_val_if_fail(name != NULL, FALSE);

  res = load_document(math_view->interface, name);
  if (res == FALSE) return FALSE;

  layout_document(math_view->interface, math_view->area->allocation.width);

  setup_adjustments(math_view);

  paint_widget(math_view);

  return TRUE;
}

void
gtk_math_view_unload(GtkMathView* math_view)
{
  g_return_if_fail(math_view != NULL);
  unload_document(math_view->interface);
}

void
gtk_math_view_dump(GtkMathView* math_view)
{
  g_return_if_fail(math_view != NULL);

  dump_document(math_view->interface);
}

GdkPixmap*
gtk_math_view_get_buffer(GtkMathView* math_view)
{
  g_return_val_if_fail(math_view != NULL, NULL);

  return math_view->pixmap;
}

void
gtk_math_view_set_adjustments(GtkMathView* math_view,
			      GtkAdjustment* hadj,
			      GtkAdjustment* vadj)
{
  g_return_if_fail(math_view != NULL);
  g_return_if_fail(hadj != NULL);
  g_return_if_fail(vadj != NULL);

  gtk_object_ref(GTK_OBJECT(hadj));
  gtk_object_ref(GTK_OBJECT(vadj));

  if (math_view->hadjustment != NULL) {
    if (math_view->hsignal != 0)
      gtk_signal_disconnect(GTK_OBJECT(math_view->hadjustment), math_view->hsignal);

    gtk_object_unref(GTK_OBJECT(math_view->hadjustment));
  }
  if (math_view->vadjustment != NULL) {
    if (math_view->vsignal != 0)
      gtk_signal_disconnect(GTK_OBJECT(math_view->vadjustment), math_view->vsignal);

    gtk_object_unref(GTK_OBJECT(math_view->vadjustment));
  }

  math_view->hadjustment = hadj;
  math_view->vadjustment = vadj;

  setup_adjustments(math_view);

  math_view->hsignal = gtk_signal_connect(GTK_OBJECT(hadj), "value_changed",
					  GTK_SIGNAL_FUNC(hadjustment_value_changed),
					  math_view);

  math_view->vsignal = gtk_signal_connect(GTK_OBJECT(vadj), "value_changed",
					  GTK_SIGNAL_FUNC(vadjustment_value_changed),
					  math_view);
}

void
gtk_math_view_clicked(GtkMathView* math_view, gpointer user_data)
{
  g_return_if_fail(math_view != NULL);
  // noop
}

void
gtk_math_view_jump(GtkMathView* math_view, const gchar* str, gpointer user_data)
{
  g_return_if_fail(math_view != NULL);
  // noop
}

GtkAdjustment*
gtk_math_view_get_hadjustment(GtkMathView* math_view)
{
  g_return_val_if_fail(math_view != NULL, NULL);

  return math_view->hadjustment;
}

GtkAdjustment*
gtk_math_view_get_vadjustment(GtkMathView* math_view)
{
  g_return_val_if_fail(math_view != NULL, NULL);

  return math_view->vadjustment;
}

GtkFrame*
gtk_math_view_get_frame(GtkMathView* math_view)
{
  g_return_val_if_fail(math_view != NULL, NULL);

  return math_view->frame != NULL ? GTK_FRAME(math_view->frame) : NULL;
}

void
gtk_math_view_set_font_size(GtkMathView* math_view, guint size)
{
  g_return_if_fail(math_view != NULL);
  g_return_if_fail(math_view->area != NULL);
  g_return_if_fail(size > 0);

  set_font_size(math_view->interface, size);
  setup_document(math_view->interface);
  layout_document(math_view->interface, math_view->area->allocation.width);
  setup_adjustments(math_view);
  paint_widget(math_view);
}

const char*
gtk_math_view_get_selection(GtkMathView* math_view)
{
  g_return_val_if_fail(math_view != NULL, NULL);

  return get_selection(math_view->interface);
}

gint
gtk_math_view_get_width(GtkMathView* math_view)
{
  g_return_val_if_fail(math_view != NULL, FALSE);
  g_return_val_if_fail(math_view->area != NULL, FALSE);

  return math_view->area->allocation.width;
}

gint
gtk_math_view_get_height(GtkMathView* math_view)
{
  g_return_val_if_fail(math_view != NULL, FALSE);
  g_return_val_if_fail(math_view->area != NULL, FALSE);

  return math_view->area->allocation.height;
}

#ifdef HELM_PROJECT
void
gtk_math_view_set_select_mode(GtkMathView* math_view, SelectMode mode)
{
  g_return_if_fail(math_view != NULL);

  reset_selection(math_view->interface);
  math_view->mode = mode;
}

SelectMode
gtk_math_view_get_select_mode(GtkMathView* math_view)
{
  g_return_val_if_fail(math_view, MATH_VIEW_SELECT_PRESENTATION);

  return math_view->mode;
}
#endif
